#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "amc1.h"

void init_AMC1(amc1_t* AMC1, uint8_t flags) {
    AMC1->ip = 0;
    AMC1->flags = flags;
}

void mem_dump(amc1_t* AMC1) {
    printf("-- MEMORY DUMP --\n");
    printf("IP: %02X\n", AMC1->ip);
    printf("FLAGS: %02X\n", AMC1->flags);
    printf("  + | 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F\n"); 
    printf(" ---+------------------------------------------------\n");
    for (int j=0; j < 256; j+=16) {
        printf(" %02X | ", j);
        for (int i=0; i < 16; i++) {
            printf("%02X ", (uint8_t)AMC1->memory[j+i]);
        }
        printf("\n");
    }
    printf("\n");
}

int8_t peek(amc1_t* AMC1, uint8_t index) {
    return AMC1->memory[index];
}

int8_t read_and_advance(amc1_t* AMC1) {
    return peek(AMC1, ++(AMC1->ip));
}

uint8_t address_at_next(amc1_t* AMC1) {
    return peek(AMC1, read_and_advance(AMC1));
}


void poke(amc1_t* AMC1, uint8_t index, int8_t value) {
    AMC1->memory[index] = value;
}


bool decode(amc1_t* AMC1) {
// Returns true if instuction is a program exit and false otherwise

#define BINARY_OP(op) \
    do { \
        int8_t a = read_and_advance(AMC1); \
        int8_t b = read_and_advance(AMC1); \
        poke(AMC1, read_and_advance(AMC1), (peek(AMC1, a) op (peek(AMC1, b)))); \
    } while (false)

#define COMPARISON(op) \
    do { \
        int8_t comp = address_at_next(AMC1); \
        if (comp op 0) { \
            AMC1->ip = read_and_advance(AMC1) - 1; \
        } else { \
            AMC1->ip++; \
        } \
    } while (false)

    uint8_t opcode = peek(AMC1, AMC1->ip);
    switch (opcode) {
        case OP_NOOP:
            printf("OP_NOOP\n");
            return false;
        case OP_INV: {
            int8_t a = read_and_advance(AMC1);
            poke(AMC1, a, ~(AMC1->memory[a]));
            return false;
        } 
        case OP_ADD: BINARY_OP(+); return false;
        case OP_SUB: BINARY_OP(-); return false;
        case OP_MUL: BINARY_OP(*); return false;
        case OP_DIV: BINARY_OP(/); return false;
        case OP_MOD: BINARY_OP(%); return false;
        case OP_AND: BINARY_OP(&); return false;
        case OP_OR:  BINARY_OP(|); return false;
        case OP_PRINT: {
            uint8_t a = read_and_advance(AMC1);
            printf("OP_PRINT: %02X\n", (uint8_t)peek(AMC1, a));
            return false;
        }
        case OP_JUMP: {
            uint8_t dest = read_and_advance(AMC1);
            printf("OP_JUMP: Jumping to memory position %02X\n", dest);
            AMC1->ip = dest - 1;
            return false;
        }
        case OP_JGZ: COMPARISON(>); return false;
        case OP_JLZ: COMPARISON(<); return false;
        case OP_JEZ: COMPARISON(==); return false;
        case OP_JNZ: COMPARISON(!=); return false;
        case OP_JGE: COMPARISON(>=); return false;
        case OP_JLE: COMPARISON(<=); return false;
        case OP_BRK: {
            mem_dump(AMC1);
            printf("Reached breakpoint at position %02X, press any key to continue", (uint8_t)AMC1->ip);
            getchar();
            return false;
        }
        case OP_EXIT:
            printf("OP_EXIT: Program exiting...\n");
            return true;
        default:
            printf("Invalid opcode: %02X. Aborting program...\n", opcode);
            return true;
    }
#undef BINARY_OP
#undef COMPARISON
}


void load_program(amc1_t* AMC1, int8_t* program, size_t prog_size) {
    if (prog_size > 256) {
        fprintf(stderr, "Program too large to load\n");
        return;
    }
    memcpy(AMC1->memory, program, prog_size);
}

void run(amc1_t* AMC1) {
    for (bool exit = false; exit != true; AMC1->ip++) {
        if (AMC1->flags & FLAG_TRAP) {
            mem_dump(AMC1);
            printf("Trap flag set. Single-step mode active. Press any key to step...\n");
            getchar();
        }
        exit = decode(AMC1);
    }
}

int main(int argc, char** argv) {
    amc1_t AMC1;
    init_AMC1(&AMC1, FLAG_TRAP);

int8_t prog1[] = {
    OP_JUMP, 0x0A,
    0x05,
    0x10,
    OP_NOOP,
    0x07,
    OP_NOOP, OP_NOOP, OP_NOOP, OP_NOOP,
    OP_ADD, 2, 3, 4,
    OP_PRINT, 4,
    OP_DIV, 4, 5, 6,
    OP_PRINT, 6,
    OP_EXIT
};
    
int8_t prog2[] = {
    OP_SUB, 10, 1, 1,
    OP_PRINT, 1,
    OP_JGZ, 1, 0,
    OP_EXIT
};

int8_t* programs[] = {prog1, prog2};

int prog_count = 2;
size_t prog_size[] = {23, 10};

    for (int i = 0; i < prog_count; i++) {
        init_AMC1(&AMC1, FLAG_TRAP);
        load_program(&AMC1, programs[i], prog_size[i]);
        printf("--- PROGRAM %d ---\n", i + 1);
        run(&AMC1);
    }
    mem_dump(&AMC1);
    return EXIT_SUCCESS;
}
